@@ -4,7 +4,9 @@ pub use nvvm::*;
 use serde::Deserialize;
 use std::{
     borrow::Borrow,
-    env, fmt,
+    env,
+    ffi::OsString,
+    fmt,
     path::{Path, PathBuf},
     process::{Command, Stdio},
 };
@@ -31,6 +33,25 @@ impl fmt::Display for CudaBuilderError {
     }
 }
 
+#[derive(Debug, Clone, Copy, PartialEq)]
+pub enum DebugInfo {
+    None,
+    LineTables,
+    // NOTE(RDambrosio016): currently unimplemented because it causes a segfault somewhere in LLVM
+    // or libnvvm. Probably the latter.
+    // Full,
+}
+
+impl DebugInfo {
+    fn into_nvvm_and_rustc_options(self) -> (String, String) {
+        match self {
+            DebugInfo::None => unreachable!(),
+            DebugInfo::LineTables => ("-generate-line-info".into(), "-Cdebuginfo=1".into()),
+            // DebugInfo::Full => ("-g".into(), "-Cdebuginfo=2".into()),
+        }
+    }
+}
+
 pub enum EmitOption {
     LlvmIr,
     Bitcode,
@@ -105,6 +126,20 @@ pub struct CudaBuilder {
     ///
     /// `false` by default.
     pub optix: bool,
+    /// Whether to override calls to [`libm`](https://docs.rs/libm/latest/libm/) with calls to libdevice intrinsics.
+    ///
+    /// Libm is used by no_std crates for functions such as sin, cos, fabs, etc. However, CUDA provides
+    /// extremely fast GPU-specific implementations of such functions through `libdevice`. Therefore, the codegen
+    /// exposes the option to automatically override any calls to libm functions with calls to libdevice functions.
+    /// However, this means the overriden functions are likely to not be deterministic, so if you rely on strict
+    /// determinism in things like `rapier`, then it may be helpful to disable such a feature.
+    ///
+    /// `true` by default.
+    pub override_libm: bool,
+    /// Whether to generate any debug info and what level of info to generate.
+    pub debug: DebugInfo,
+    /// Additional arguments passed to cargo during `cargo build`.
+    pub build_args: Vec<String>,
 }
 
 impl CudaBuilder {
@@ -123,9 +158,25 @@ impl CudaBuilder {
             fma_contraction: true,
             emit: None,
             optix: false,
+            override_libm: true,
+            debug: DebugInfo::None,
+            build_args: vec![],
         }
     }
 
+    /// Additional arguments passed to cargo during `cargo build`.
+    pub fn build_args(mut self, args: &[impl AsRef<str>]) -> Self {
+        self.build_args
+            .extend(args.iter().map(|s| s.as_ref().to_owned()));
+        self
+    }
+
+    /// Whether to generate any debug info and what level of info to generate.
+    pub fn debug(mut self, debug: DebugInfo) -> Self {
+        self.debug = debug;
+        self
+    }
+
     /// Whether to compile the gpu crate for release.
     pub fn release(mut self, release: bool) -> Self {
         self.release = release;
@@ -231,14 +282,25 @@ impl CudaBuilder {
         self
     }
 
+    /// Whether to override calls to [`libm`](https://docs.rs/libm/latest/libm/) with calls to libdevice intrinsics.
+    ///
+    /// Libm is used by no_std crates for functions such as sin, cos, fabs, etc. However, CUDA provides
+    /// extremely fast GPU-specific implementations of such functions through `libdevice`. Therefore, the codegen
+    /// exposes the option to automatically override any calls to libm functions with calls to libdevice functions.
+    /// However, this means the overriden functions are likely to not be deterministic, so if you rely on strict
+    /// determinism in things like `rapier`, then it may be helpful to disable such a feature.
+    pub fn override_libm(mut self, override_libm: bool) -> Self {
+        self.override_libm = override_libm;
+        self
+    }
+
     /// Runs rustc to build the codegen and codegens the gpu crate, returning the path of the final
     /// ptx file. If [`ptx_file_copy_path`](Self::ptx_file_copy_path) is set, this returns the copied path.
     pub fn build(self) -> Result<PathBuf, CudaBuilderError> {
         println!("cargo:rerun-if-changed={}", self.path_to_crate.display());
         let path = invoke_rustc(&self)?;
         if let Some(copy_path) = self.ptx_file_copy_path {
-            std::fs::copy(path, &copy_path)
-                .map_err(|x| CudaBuilderError::FailedToCopyPtxFile(x))?;
+            std::fs::copy(path, &copy_path).map_err(CudaBuilderError::FailedToCopyPtxFile)?;
             Ok(copy_path)
         } else {
             Ok(path)
@@ -279,15 +341,19 @@ fn find_rustc_codegen_nvvm() -> PathBuf {
     panic!("Could not find {} in library path", filename);
 }
 
-fn add_libnvvm_to_path() {
+fn get_new_path_var() -> OsString {
     let split_paths = env::var_os(dylib_path_envvar()).unwrap_or_default();
     let mut paths = env::split_paths(&split_paths).collect::<Vec<_>>();
-    let libnvvm_path = PathBuf::from(find_cuda_helper::find_cuda_root().unwrap())
-        .join("nvvm")
-        .join("bin");
-    paths.push(libnvvm_path);
-    let joined = env::join_paths(&paths).expect("Failed to join paths for PATH");
-    env::set_var(dylib_path_envvar(), joined);
+    let possible_paths = if cfg!(target_os = "windows") {
+        vec![find_cuda_helper::find_cuda_root()
+            .unwrap()
+            .join("nvvm")
+            .join("bin")]
+    } else {
+        find_cuda_helper::find_cuda_lib_dirs()
+    };
+    paths.extend(possible_paths);
+    env::join_paths(&paths).expect("Failed to join paths for PATH")
 }
 
 /// Joins strings together while ensuring none of the strings contain the separator.
@@ -309,7 +375,7 @@ fn invoke_rustc(builder: &CudaBuilder) -> Result<PathBuf, CudaBuilderError> {
     // on what this does
     let rustc_codegen_nvvm = find_rustc_codegen_nvvm();
 
-    add_libnvvm_to_path();
+    let new_path = get_new_path_var();
 
     let mut rustflags = vec![format!(
         "-Zcodegen-backend={}",
@@ -324,9 +390,7 @@ fn invoke_rustc(builder: &CudaBuilder) -> Result<PathBuf, CudaBuilderError> {
         rustflags.push(format!("--emit={}", string));
     }
 
-    let mut llvm_args = vec![];
-
-    llvm_args.push(NvvmOption::Arch(builder.arch).to_string());
+    let mut llvm_args = vec![NvvmOption::Arch(builder.arch).to_string()];
 
     if !builder.nvvm_opts {
         llvm_args.push("-opt=0".to_string());
@@ -348,6 +412,16 @@ fn invoke_rustc(builder: &CudaBuilder) -> Result<PathBuf, CudaBuilderError> {
         llvm_args.push("-fma=0".to_string());
     }
 
+    if builder.override_libm {
+        llvm_args.push("--override-libm".to_string());
+    }
+
+    if builder.debug != DebugInfo::None {
+        let (nvvm_flag, rustc_flag) = builder.debug.into_nvvm_and_rustc_options();
+        llvm_args.push(nvvm_flag);
+        rustflags.push(rustc_flag);
+    }
+
     let llvm_args = llvm_args.join(" ");
     if !llvm_args.is_empty() {
         rustflags.push(["-Cllvm-args=", &llvm_args].concat());
@@ -369,10 +443,18 @@ fn invoke_rustc(builder: &CudaBuilder) -> Result<PathBuf, CudaBuilderError> {
         target,
     ]);
 
+    cargo.args(&builder.build_args);
+
+    cargo.env(dylib_path_envvar(), new_path);
+
     if builder.release {
         cargo.arg("--release");
     }
 
+    // TODO(RDambrosio016): Remove this once we can get meaningful error messages in panic to work.
+    // for now we enable it to remove some useless indirect calls in the ptx.
+    cargo.arg("-Zbuild-std-features=panic_immediate_abort");
+
     if builder.optix {
         cargo.arg("-Zbuild-std-features=panic_immediate_abort");
         cargo.arg("-Zunstable-options");
@@ -401,6 +483,9 @@ fn invoke_rustc(builder: &CudaBuilder) -> Result<PathBuf, CudaBuilderError> {
         }
     }
 
+    let arch = format!("{:?}0", builder.arch);
+    cargo.env("CUDA_ARCH", arch.strip_prefix("Compute").unwrap());
+
     let cargo_encoded_rustflags = join_checking_for_separators(rustflags, "\x1f");
 
     let build = cargo
